{
    $Id: objects.pp,v 1.37 2000/04/07 21:10:35 pierre Exp $
    This file is part of the Free Pascal run time library.
    Copyright (c) 1999-2000 by the Free Pascal development team.

    Objects.pas clone for Free Pascal

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}

{************[ SOURCE FILE OF FREE VISION ]****************}
{                                                          }
{    System independent clone of objects.pas               }
{                                                          }
{    Interface Copyright (c) 1992 Borland International    }
{                                                          }
{    Parts Copyright (c) 1999-2000 by Florian Klaempfl       }
{    fnklaemp@cip.ft.uni-erlangen.de                       }
{                                                          }
{    Parts Copyright (c) 1999-2000 by Frank ZAGO                }
{    zago@ecoledoc.ipc.fr                                  }
{                                                          }
{    Parts Copyright (c) 1999-2000 by MH Spiegel                }
{                                                          }
{    Parts Copyright (c) 1996, 1999-2000 by Leon de Boer        }
{    ldeboer@ibm.net                                       }
{                                                          }
{    Free Vision project coordinator Balazs Scheidler      }
{    bazsi@tas.vein.hu                                     }
{                                                          }
unit Objects;

{$I platform.inc}

{<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>}
interface
{<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>}

{==== Select assembler ==============================================}
{$IFDEF CPU86}
{$ASMMODE ATT}
{$ENDIF}

{$IFDEF CPU68}
{$ASMMODE MOT}
{$ENDIF}

{==== Compiler directives ===========================================}
{$H-} { No ansistrings }
{$E+} { Emulation is on }
{$X+} { Extended syntax is ok }
{$R-} { Disable range checking }
{$IFNDEF Linux}
{$S-} { Disable Stack Checking }
{$ENDIF}
{$I-} { Disable IO Checking }
{$Q-} { Disable Overflow Checking }
{$V-} { Turn off strict VAR strings }
{====================================================================}

{***************************************************************************}
{                             PUBLIC CONSTANTS                              }
{***************************************************************************}

const
  MaxAvail = MaxInt;

  {---------------------------------------------------------------------------}
  {                          STREAM ERROR STATE MASKS                         }
  {---------------------------------------------------------------------------}
const
  stOk = 0; { No stream error }
  stError = -1; { Access error }
  stInitError = -2; { Initialize error }
  stReadError = -3; { Stream read error }
  stWriteError = -4; { Stream write error }
  stGetError = -5; { Get object error }
  stPutError = -6; { Put object error }
  stSeekError = -7; { Seek error in stream }
  stOpenError = -8; { Error opening stream }

  {---------------------------------------------------------------------------}
  {                        STREAM ACCESS MODE CONSTANTS                       }
  {---------------------------------------------------------------------------}
const
  stCreate = $3C00; { Create new file }
  stOpenRead = $3D00; { Read access only }
  stOpenWrite = $3D01; { Write access only }
  stOpen = $3D02; { Read/write access }

  {---------------------------------------------------------------------------}
  {                          TCollection ERROR CODES                          }
  {---------------------------------------------------------------------------}
const
  coIndexError = -1; { Index out of range }
  coOverflow = -2; { Overflow }

  {---------------------------------------------------------------------------}
  {         VMT HEADER CONSTANT - HOPEFULLY WE CAN DROP THIS LATER            }
  {---------------------------------------------------------------------------}
const
  vmtHeaderSize = 8; { VMT header size }

const
  {---------------------------------------------------------------------------}
  {                            MAXIUM DATA SIZES                              }
  {---------------------------------------------------------------------------}
  {$IFDEF FPC}
  MaxBytes = 128 * 1024 * 1024; { Maximum data size }
  {$ELSE}
  MaxBytes = 16384;
  {$ENDIF}
  MaxWords = MaxBytes div SizeOf(Word); { Max word data size }
  MaxPtrs = MaxBytes div SizeOf(Pointer); { Max ptr data size }
  MaxCollectionSize = MaxBytes div SizeOf(Pointer); { Max collection size }
  MaxTPCompatibleCollectionSize = 65520 div 4;

  {***************************************************************************}
  {                          PUBLIC TYPE DEFINITIONS                          }
  {***************************************************************************}

type
  DWORD = Cardinal;

  {---------------------------------------------------------------------------}
  {                               CHARACTER SET                               }
  {---------------------------------------------------------------------------}
type
  TCharSet = set of Char; { Character set }
  PCharSet = ^TCharSet; { Character set ptr }

  {---------------------------------------------------------------------------}
  {                               GENERAL ARRAYS                              }
  {---------------------------------------------------------------------------}
type
  TByteArray = array[0..MaxBytes - 1] of Byte; { Byte array }
  PByteArray = ^TByteArray; { Byte array pointer }

  TWordArray = array[0..MaxWords - 1] of Word; { Word array }
  PWordArray = ^TWordArray; { Word array pointer }

  TPointerArray = array[0..MaxPtrs - 1] of Pointer; { Pointer array }
  PPointerArray = ^TPointerArray; { Pointer array ptr }

  {---------------------------------------------------------------------------}
  {                             POINTER TO STRING                             }
  {---------------------------------------------------------------------------}
type
  PString = ^string; { String pointer }

  {---------------------------------------------------------------------------}
  {                    OS dependent File type / consts                        }
  {---------------------------------------------------------------------------}
  {$IFDEF GO32V1}
type
  FNameStr = string[79];
  THandle = Integer;
const
  MaxReadBytes = $FFFE;
  invalidhandle = -1;
  {$ENDIF}
  {$IFDEF GO32V2}
type
  FNameStr = string;
  THandle = Integer;
const
  MaxReadBytes = $FFFE;
  invalidhandle = -1;
  {$ENDIF}
  {$IFDEF Win32}
type
  FNameStr = string;
  THandle = LongInt;
const
  MaxReadBytes = $FFFE;
  invalidhandle = -1;
  {$ENDIF}
  {$IFDEF OS2}
type
  FNameStr = string;
  THandle = Word;
const
  MaxReadBytes = $7FFFFFFF;
  invalidhandle = $FFFF;
  {$ENDIF}
  {$IFDEF LINUX}
type
  FNameStr = string;
  { values are words, though the OS calls return 32-bit values }
  { to check (CEC)                                             }
  THandle = LongInt;
const
  MaxReadBytes = $7FFFFFFF;
  invalidhandle = -1;
  {$ENDIF}
  {$IFDEF AMIGA}
type
  FNameStr = string;
  THandle = LongInt;
const
  MaxReadBytes = $FFFE;
  invalidhandle = -1;
  {$ENDIF}
  {$IFDEF ATARI}
type
  FNameStr = string[79];
  THandle = Integer;
const
  MaxReadBytes = $FFFE;
  invalidhandle = -1;
  {$ENDIF}
  {$IFDEF MAC}
type
  FNameStr = string;
  THandle = Integer;
const
  MaxReadBytes = $FFFE;
  invalidhandle = -1;
  {$ENDIF}

  {---------------------------------------------------------------------------}
  {                            DOS ASCIIZ FILENAME                            }
  {---------------------------------------------------------------------------}
type
  AsciiZ = array[0..255] of Char; { Filename array }

  {---------------------------------------------------------------------------}
  {                        BIT SWITCHED TYPE CONSTANTS                        }
  {---------------------------------------------------------------------------}
type
  Sw_Word = LongInt; { Long Word now }
  Sw_Integer = LongInt; { Long integer now }

  {***************************************************************************}
  {                        PUBLIC RECORD DEFINITIONS                          }
  {***************************************************************************}

  {---------------------------------------------------------------------------}
  {                          TYPE CONVERSION RECORDS                          }
  {---------------------------------------------------------------------------}
type
  WordRec = packed record
    Lo, Hi: Byte; { Word to bytes }
  end;

  LongRec = packed record
    Lo, Hi: Word; { LongInt to words }
  end;

  PtrRec = packed record
    Ofs, Seg: Word; { Pointer to words }
  end;

  {---------------------------------------------------------------------------}
  {                  TStreamRec RECORD - STREAM OBJECT RECORD                 }
  {---------------------------------------------------------------------------}
type
  PStreamRec = ^TStreamRec; { Stream record ptr }
  TStreamRec = packed record
    ObjType: Sw_Word; { Object type id }
    VmtLink: Pointer; { VMT link }
    Load: Pointer; { Object load code }
    STORE: Pointer; { Object store code }
    Next: PStreamRec; { Next stream record }
  end;

  {***************************************************************************}
  {                        PUBLIC OBJECT DEFINITIONS                          }
  {***************************************************************************}

{---------------------------------------------------------------------------}
{                  TObject OBJECT - BASE ANCESTOR OBJECT                    }
{---------------------------------------------------------------------------}
type
  TObject = object
    constructor Init;
    procedure Free;
    function Is_Object(p: Pointer): Boolean;
    destructor Done; virtual;
  end;
  PObject = ^TObject;

  { ******************************* REMARK ****************************** }
  {  Two new virtual methods have been added to the object in the form of }
  {  Close and Open. The main use here is in the Disk Based Descendants   }
  {  the calls open and close the given file so these objects can be      }
  {  used like standard files. Two new fields have also been added to     }
  {  speed up seeks on descendants. All existing code will compile and    }
  {  work completely normally oblivious to these new methods and fields.  }
  { ****************************** END REMARK *** Leon de Boer, 15May96 * }

  {---------------------------------------------------------------------------}
  {                 TStream OBJECT - STREAM ANCESTOR OBJECT                   }
  {---------------------------------------------------------------------------}
type
  TStream = object(TObject)
    Status: Integer; { Stream status }
    ErrorInfo: Integer; { Stream error info }
    StreamSize: LongInt; { Stream current size }
    Position: LongInt; { Current position }
    TPCompatible: Boolean;
    constructor Init;
    // RJ function Get: PObject;
    function StrRead: PChar;
    function GetPos: LongInt; virtual;
    function GetSize: LongInt; virtual;
    function ReadStr: PString;
    procedure Open(OpenMode: Word); virtual;
    procedure Close; virtual;
    procedure Reset;
    procedure Flush; virtual;
    procedure Truncate; virtual;
    // RJ procedure Put (p: PObject);
    procedure StrWrite(p: PChar);
    procedure WriteStr(p: PString);
    procedure Seek(Pos: LongInt); virtual;
    procedure Error(Code, Info: Integer); virtual;
    procedure Read(var Buf; Count: Sw_Word); virtual;
    procedure Write(var Buf; Count: Sw_Word); virtual;
    procedure CopyFrom(var s: TStream; Count: LongInt);
  end;
  PStream = ^TStream;

  { ******************************* REMARK ****************************** }
  {  All the changes here should be completely transparent to existing    }
  {  code. Basically the memory blocks do not have to be base segments    }
  {  but this means our list becomes memory blocks rather than segments.  }
  {  The stream will also expand like the other standard streams!!        }
  { ****************************** END REMARK *** Leon de Boer, 19May96 * }

type
  TItemList = array[0..MaxCollectionSize - 1] of Pointer;
  PItemList = ^TItemList;

  { ******************************* REMARK ****************************** }
  {    The changes here look worse than they are. The Sw_Integer simply   }
  {  switches between Integers and LongInts if switched between 16 and 32 }
  {  bit code. All existing code will compile without any changes.        }
  { ****************************** END REMARK *** Leon de Boer, 10May96 * }

  {---------------------------------------------------------------------------}
  {              TCollection OBJECT - COLLECTION ANCESTOR OBJECT              }
  {---------------------------------------------------------------------------}
  TCollection = object(TObject)
    Items: PItemList; { Item list pointer }
    Count: Sw_Integer; { Item count }
    Limit: Sw_Integer; { Item limit count }
    Delta: Sw_Integer; { Inc delta size }
    constructor Init(ALimit, ADelta: Sw_Integer);
    // RJ constructor Load (var s: TStream);
    destructor Done; virtual;
    function At(Index: Sw_Integer): Pointer;
    function IndexOf(Item: Pointer): Sw_Integer; virtual;
    // RJ function GetItem (var s: TStream): Pointer; virtual;
    // RJ function LastThat (test: Pointer): Pointer;
    // RJ function FirstThat (test: Pointer): Pointer;
    procedure Pack;
    procedure FreeAll;
    procedure DeleteAll;
    procedure Free(Item: Pointer);
    procedure Insert(Item: Pointer); virtual;
    procedure Delete(Item: Pointer);
    procedure AtFree(Index: Sw_Integer);
    procedure FreeItem(Item: Pointer); virtual;
    procedure AtDelete(Index: Sw_Integer);
    // RJ procedure forEach (Action: Pointer);
    procedure SetLimit(ALimit: Sw_Integer); virtual;
    procedure Error(Code, Info: Integer); virtual;
    procedure AtPut(Index: Sw_Integer; Item: Pointer);
    procedure AtInsert(Index: Sw_Integer; Item: Pointer);
    // RJ procedure STORE (var s: TStream);
    // RJ procedure PutItem (var s: TStream; Item: Pointer); virtual;
  end;
  PCollection = ^TCollection;

  {---------------------------------------------------------------------------}
  {          TSortedCollection OBJECT - SORTED COLLECTION ANCESTOR            }
  {---------------------------------------------------------------------------}
type
  TSortedCollection = object(TCollection)
    Duplicates: Boolean; { Duplicates flag }
    constructor Init(ALimit, ADelta: Sw_Integer);
    // RJ constructor Load (var s: TStream);
    function KeyOf(Item: Pointer): Pointer; virtual;
    function IndexOf(Item: Pointer): Sw_Integer; virtual;
    function Compare(Key1, Key2: Pointer): Sw_Integer; virtual;
    function Search(Key: Pointer; var Index: Sw_Integer): Boolean; virtual;
    procedure Insert(Item: Pointer); virtual;
    // RJ procedure STORE (var s: TStream);
  end;
  PSortedCollection = ^TSortedCollection;

  {---------------------------------------------------------------------------}
  {           TStringCollection OBJECT - STRING COLLECTION OBJECT             }
  {---------------------------------------------------------------------------}
type
  TStringCollection = object(TSortedCollection)
    function GetItem(var s: TStream): Pointer; virtual;
    function Compare(Key1, Key2: Pointer): Sw_Integer; virtual;
    procedure FreeItem(Item: Pointer); virtual;
    procedure PutItem(var s: TStream; Item: Pointer); virtual;
  end;
  PStringCollection = ^TStringCollection;

  {---------------------------------------------------------------------------}
  {        TUnSortedStringCollection - UNSORTED STRING COLLECTION OBJECT         }
  {---------------------------------------------------------------------------}
type
  TUnSortedStringCollection = object(TStringCollection)
    procedure Insert(Item: Pointer); virtual;
  end;
  PUnSortedStringCollection = ^TUnSortedStringCollection;

  {---------------------------------------------------------------------------}
  {                 TResourceFile OBJECT - RESOURCE FILE OBJECT               }
  {---------------------------------------------------------------------------}

(*
type
  TResourceFile = object (TObject)
    Stream: PStream;                    { File as a stream }
    Modified: Boolean;                  { Modified flag }
    constructor Init (AStream: PStream);
    destructor Done; virtual;
    function Count: Sw_Integer;
    function KeyAt (i: Sw_Integer): string;
    function Get (Key: string): PObject;
    function SwitchTo (AStream: PStream; Pack: Boolean): PStream;
    procedure Flush;
    procedure Delete (Key: string);
    procedure Put (Item: PObject; Key: string);
  private
    BasePos: LongInt;                   { Base position }
    IndexPos: LongInt;                  { Index position }
    Index: TResourceCollection;         { Index collection }
  end;
  PResourceFile = ^TResourceFile;
*)

type
  TStrIndexRec = packed record
    Key, Count, Offset: Word;
  end;

  TStrIndex = array[0..9999] of TStrIndexRec;
  PStrIndex = ^TStrIndex;

  {---------------------------------------------------------------------------}
  {                 TStringList OBJECT - STRING LIST OBJECT                   }
  {---------------------------------------------------------------------------}
  TStringList = object(TObject)
    constructor Load(var s: TStream);
    destructor Done; virtual;
    function Get(Key: Sw_Word): string;
  private
    Stream: PStream;
    BasePos: LongInt;
    IndexSize: Sw_Word;
      Index: PStrIndex;
    procedure ReadStr(var s: string; Offset, Skip: Sw_Word);
  end;
  PStringList = ^TStringList;

  {---------------------------------------------------------------------------}
  {                 TStrListMaker OBJECT - RESOURCE FILE OBJECT               }
  {---------------------------------------------------------------------------}
type
  TStrListMaker = object(TObject)
    constructor Init(AStrSize, AIndexSize: Sw_Word);
    destructor Done; virtual;
    procedure Put(Key: Sw_Word; s: string);
    procedure STORE(var s: TStream);
  private
    StrPos: Sw_Word;
    StrSize: Sw_Word;
    strings: PByteArray;
    IndexPos: Sw_Word;
    IndexSize: Sw_Word;
      Index: PStrIndex;
    cur: TStrIndexRec;
    procedure CloseCurrent;
  end;
  PStrListMaker = ^TStrListMaker;

  {***************************************************************************}
  {                            INTERFACE ROUTINES                             }
  {***************************************************************************}

  {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  {                    DYNAMIC STRING INTERFACE ROUTINES                      }
  {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  {-NewStr-------------------------------------------------------------
  Allocates a dynamic string into memory. If S is nil, NewStr returns
  a nil pointer, otherwise NewStr allocates Length(S)+1 bytes of memory
  containing a copy of S, and returns a pointer to the string.
  12Jun96 LdB
  ---------------------------------------------------------------------}
function NewStr(const s: string): PString;

{-DisposeStr---------------------------------------------------------
Disposes of a PString allocated by the function NewStr.
12Jun96 LdB
---------------------------------------------------------------------}
procedure DisposeStr(p: PString);

procedure SetStr(var p: PString; const s: string);

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                        STREAM INTERFACE ROUTINES                          }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

{-Abstract-----------------------------------------------------------
Terminates program with a run-time error 211. When implementing
an abstract object type, call Abstract in those virtual methods that
must be overridden in descendant types. This ensures that any
attempt to use instances of the abstract object type will fail.
12Jun96 LdB
---------------------------------------------------------------------}
procedure Abstract;

{-RegisterObjects----------------------------------------------------
Registers the three standard objects TCollection, TStringCollection
and TStrCollection.
02Sep97 LdB
---------------------------------------------------------------------}
procedure RegisterObjects;

{-RegisterType-------------------------------------------------------
Registers the given object type with Free Vision's streams, creating
a list of known objects. Streams can only store and return these known
object types. Each registered object needs a unique stream registration
record, of type TStreamRec.
02Sep97 LdB
---------------------------------------------------------------------}
procedure RegisterType(var s: TStreamRec);

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                    GENERAL FUNCTION INTERFACE ROUTINES                    }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

{-LongMul------------------------------------------------------------
Returns the long integer value of X * Y integer values.
04Sep97 LdB
---------------------------------------------------------------------}
function LongMul(X, y: Integer): LongInt;

{-LongDiv------------------------------------------------------------
Returns the integer value of long integer X divided by integer Y.
04Sep97 LdB
---------------------------------------------------------------------}
function LongDiv(X: LongInt; y: Integer): Integer;

{***************************************************************************}
{                         PUBLIC INITIALIZED VARIABLES                      }
{***************************************************************************}

const
  {---------------------------------------------------------------------------}
  {              INITIALIZED DOS/DPMI/WIN/OS2 PUBLIC VARIABLES                }
  {---------------------------------------------------------------------------}
  StreamError: Pointer = nil; { Stream error ptr }
  DosStreamError: Word = $0; { Dos stream error }
  DefaultTPCompatible: Boolean = False;

  {---------------------------------------------------------------------------}
  {                        STREAM REGISTRATION RECORDS                        }
  {---------------------------------------------------------------------------}

  {CONST
     RCollection: TStreamRec = (
       ObjType: 50;
       VmtLink: Ofs(TypeOf(TCollection)^);
       Load: @TCollection.Load;
       Store: @TCollection.Store);

     RStringCollection: TStreamRec = (
       ObjType: 51;
       VmtLink: Ofs(TypeOf(TStringCollection)^);
       Load: @TStringCollection.Load;
       Store: @TStringCollection.Store);

     RStrCollection: TStreamRec = (
       ObjType: 69;
       VmtLink: Ofs(TypeOf(TStrCollection)^);
       Load:    @TStrCollection.Load;
       Store:   @TStrCollection.Store);

     RStringList: TStreamRec = (
       ObjType: 52;
       VmtLink: Ofs(TypeOf(TStringList)^);
       Load: @TStringList.Load;
       Store: Nil);

     RStrListMaker: TStreamRec = (
       ObjType: 52;
       VmtLink: Ofs(TypeOf(TStrListMaker)^);
       Load: Nil;
       Store: @TStrListMaker.Store); }

  {<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>}
implementation
{<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>}

{***************************************************************************}
{                      HELPER ROUTINES FOR CALLING                          }
{***************************************************************************}

type
  FramePointer = Pointer;
  PointerLocal = function(_EBP: FramePointer; Param1: Pointer): Pointer;
  PointerConstructor = function(VMT: Pointer; Obj: Pointer; Param1: Pointer): Pointer;
  PointerMethod = function(Obj: Pointer; Param1: Pointer): Pointer;

  (*
  function PreviousFramePointer: FramePointer; assembler;
  asm
    {$ifdef i386}
        movl (%ebp), %eax
    {$endif}
    {$ifdef m68k}
        move.l a6,d0
    {$endif}
  // end ['EAX'];
  end;

  function CallPointerConstructor (Ctor: Pointer; Obj: Pointer; VMT: Pointer; Param1: Pointer): Pointer;
  begin
    asm
  {$ifdef i386}
          movl Obj, %esi
  {$endif}
  {$ifdef m68k}
          move.l Obj, a5
  {$endif}
    end;
    CallPointerConstructor := PointerConstructor (Ctor) (VMT, Obj, Param1)
  end;

  function CallPointerMethod (method: Pointer; Obj: Pointer; Param1: Pointer): Pointer;
  begin
    asm
  {$ifdef i386}
          movl Obj, %esi
  {$endif}
  {$ifdef m68k}
          move.l Obj, a5
  {$endif}
    end;
    CallPointerMethod := PointerMethod (method) (Obj, Param1)
  end;
  *)

function CallPointerLocal(func: Pointer; Frame: FramePointer; Param1: Pointer): Pointer;
begin
  CallPointerLocal := PointerLocal(func)(Frame, Param1)
end;

{***************************************************************************}
{                      PRIVATE INITIALIZED VARIABLES                        }
{***************************************************************************}

{---------------------------------------------------------------------------}
{               INITIALIZED DOS/DPMI/WIN/OS2 PRIVATE VARIABLES              }
{---------------------------------------------------------------------------}
const
  StreamTypes: PStreamRec = nil; { Stream types reg }

  {***************************************************************************}
  {                          PRIVATE INTERNAL ROUTINES                        }
  {***************************************************************************}

  {$I objinc.inc}

  {---------------------------------------------------------------------------}
  {  RegisterError -> Platforms DOS/DPMI/WINDOWS/OS2 - Checked 12Jun96 LdB    }
  {---------------------------------------------------------------------------}
procedure RegisterError;
begin
  RunError(212); { Register error }
end;

{***************************************************************************}
{                               OBJECT METHODS                              }
{***************************************************************************}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                           TObject OBJECT METHODS                          }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

type
  DummyObject = object(TObject) { Internal object }
    Data: record end; { Helps size VMT link }
  end;

  { ******************************* REMARK ****************************** }
  { I Prefer this code because it self sizes VMT link rather than using a }
  { fixed record structure thus it should work on all compilers without a }
  { specific record to match each compiler.                               }
  { ****************************** END REMARK *** Leon de Boer, 10May96 * }

  {--TObject------------------------------------------------------------------}
  {  Init -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                 }
  {---------------------------------------------------------------------------}
constructor TObject.Init;
var
  LinkSize: LongInt; Dummy: DummyObject;
begin
  LinkSize := LongInt(@Dummy.Data) - LongInt(@Dummy); { Calc VMT link size }
  FillChar(Pointer(LongInt(@Self) + LinkSize)^,
    SizeOf(Self) - LinkSize, #0); { Clear data fields }
end;

{--TObject------------------------------------------------------------------}
{  Free -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                 }
{---------------------------------------------------------------------------}
procedure TObject.Free;
begin
  Dispose(PObject(@Self), Done); { Dispose of self }
end;

{--TObject------------------------------------------------------------------}
{  Is_Object -> Platforms DOS/DPMI/WIN/OS2 - Checked 5Mar00 DM              }
{---------------------------------------------------------------------------}
function TObject.Is_Object(p: Pointer): Boolean;
type
  PVMT = ^VMT;
  VMT = record
    Size, NegSize: LongInt;
    ParentLink: PVMT;
  end;
var
  sp: ^PVMT; Q: PVMT;
begin
  sp := @Self;
  Q := sp^;
  Is_Object := False;
  while Q <> nil do
    begin
      if Q = p then
        begin
          Is_Object := True;
          Break;
        end;
      Q := Q^.ParentLink;
    end;
end;

{--TObject------------------------------------------------------------------}
{  Done -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                 }
{---------------------------------------------------------------------------}
destructor TObject.Done;
begin { Abstract method }
end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                           TStream OBJECT METHODS                          }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

constructor TStream.Init;
begin
  TPCompatible := DefaultTPCompatible;
end;

{--TStream------------------------------------------------------------------}
{  Get -> Platforms DOS/DPMI/WIN/OS2 - Checked 02Sep97 LdB                  }
{---------------------------------------------------------------------------}

(* RJ
function TStream.Get: PObject;
var
  ObjType           : Sw_Word; p: PStreamRec; ObjTypeWord: Word;
begin
  if TPCompatible then
    begin
      { Read 16-bit word for TP compatibility. }
      Read (ObjTypeWord, SizeOf (ObjTypeWord));
      ObjType := ObjTypeWord
    end
  else
    Read (ObjType, SizeOf (ObjType));   { Read object type }
  if (ObjType <> 0) then
    begin                               { Object registered }
      p := StreamTypes;                 { Current reg list }
      while (p <> nil) and (p^.ObjType <> ObjType) { Find object type OR }
      do p := p^.Next;                  { Find end of chain }
      if (p = nil) then
        begin                           { Not registered }
          Error (stGetError, ObjType);  { Obj not registered }
          Get := nil;                   { Return nil pointer }
        end else
        Get := PObject (
          CallPointerConstructor (p^.Load, nil, p^.VmtLink, @Self)) { Call constructor }
    end else Get := nil;                { Return nil pointer }
end;

*)

{--TStream------------------------------------------------------------------}
{  StrRead -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB              }
{---------------------------------------------------------------------------}
function TStream.StrRead: PChar;
var
  l: Word; p: PChar;
begin
  Read(l, SizeOf(l)); { Read length }
  if (l = 0) then
    StrRead := nil
  else
    begin { Check for empty }
      GetMem(p, l + 1); { Allocate memory }
      if (p <> nil) then
        begin { Check allocate okay }
          Read(p[0], l); { Read the data }
          p[l] := #0; { Terminate with #0 }
        end;
      StrRead := p; { Return PChar }
    end;
end;

{--TStream------------------------------------------------------------------}
{  ReadStr -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB              }
{---------------------------------------------------------------------------}
function TStream.ReadStr: PString;
var
  l: Byte; p: PString;
begin
  Read(l, 1); { Read string length }
  if (l > 0) then
    begin
      GetMem(p, l + 1); { Allocate memory }
      if (p <> nil) then
        begin { Check allocate okay }
          p^[0] := Char(l); { Hold length }
          Read(p^[1], l); { Read string data }
        end;
      ReadStr := p; { Return string ptr }
    end
  else
    ReadStr := nil;
end;

{--TStream------------------------------------------------------------------}
{  GetPos -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB               }
{---------------------------------------------------------------------------}
function TStream.GetPos: LongInt;
begin
  if (Status = stOk) then
    GetPos := Position { Return position }
  else
    GetPos := -1; { Stream in error }
end;

{--TStream------------------------------------------------------------------}
{  GetSize -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB              }
{---------------------------------------------------------------------------}
function TStream.GetSize: LongInt;
begin
  if (Status = stOk) then
    GetSize := StreamSize { Return stream size }
  else
    GetSize := -1; { Stream in error }
end;

{--TStream------------------------------------------------------------------}
{  Close -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                }
{---------------------------------------------------------------------------}
procedure TStream.Close;
begin { Abstract method }
end;

{--TStream------------------------------------------------------------------}
{  Reset -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                }
{---------------------------------------------------------------------------}
procedure TStream.Reset;
begin
  Status := 0; { Clear status }
  ErrorInfo := 0; { Clear error info }
end;

{--TStream------------------------------------------------------------------}
{  Flush -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                }
{---------------------------------------------------------------------------}
procedure TStream.Flush;
begin { Abstract method }
end;

{--TStream------------------------------------------------------------------}
{  Truncate -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB             }
{---------------------------------------------------------------------------}
procedure TStream.Truncate;
begin
  Abstract; { Abstract error }
end;

{--TStream------------------------------------------------------------------}
{  Put -> Platforms DOS/DPMI/WIN/OS2 - Checked 02Sep97 LdB                  }
{---------------------------------------------------------------------------}

(* RJ
procedure TStream.Put (p: PObject);
var
  ObjType           : Sw_Word; Link: Pointer; Q: PStreamRec; VmtPtr: ^Pointer;
  ObjTypeWord       : Word;
begin
  VmtPtr := Pointer (p);                { Xfer object to ptr }
  if Assigned (VmtPtr) then
    Link := VmtPtr^                     { VMT link }
  else
    Link := nil;
  ObjType := 0;                         { Set objtype to zero }
  if (p <> nil) and (Link <> nil) then
    begin                               { We have a VMT link }
      Q := StreamTypes;                 { Current reg list }
      while (Q <> nil) and (Q^.VmtLink <> Link) { Find link match OR }
      do Q := Q^.Next;                  { Find end of chain }
      if (Q = nil) then
        begin                           { End of chain found }
          Error (stPutError, 0);        { Not registered error }
          Exit;                         { Now exit }
        end else ObjType := Q^.ObjType; { Update object type }
    end;
  if TPCompatible then
    begin
      ObjTypeWord := ObjType;
      Write (ObjTypeWord, SizeOf (ObjTypeWord))
    end
  else
    Write (ObjType, SizeOf (ObjType));  { Write object type }
  if (ObjType <> 0) then                { Registered object }
    CallPointerMethod (Q^.STORE, p, @Self);
end;

*)

{--TStream------------------------------------------------------------------}
{  Seek -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                 }
{---------------------------------------------------------------------------}
procedure TStream.Seek(Pos: LongInt);
begin
  if (Status = stOk) then
    begin { Check status }
      if (Pos < 0) then Pos := 0; { Remove negatives }
      if (Pos <= StreamSize) then
        Position := Pos { If valid set pos }
      else
        Error(stSeekError, Pos); { Position error }
    end;
end;

{--TStream------------------------------------------------------------------}
{  StrWrite -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB             }
{---------------------------------------------------------------------------}
procedure TStream.StrWrite(p: PChar);
var
  l: Word; Q: PByteArray;
begin
  l := 0; { Preset zero size }
  Q := PByteArray(p); { Transfer type }
  if (Q <> nil) then
    while (Q^[l] <> 0) do
      Inc(l); { PChar length }
  Write(l, SizeOf(l)); { Store length }
  if (p <> nil) then Write(p[0], l); { Write data }
end;

{--TStream------------------------------------------------------------------}
{  WriteStr -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB             }
{---------------------------------------------------------------------------}
procedure TStream.WriteStr(p: PString);
const
  Empty: string[1] = '';
begin
  if (p <> nil) then
    Write(p^, Length(p^) + 1) { Write string }
  else
    Write(Empty, 1); { Write empty string }
end;

{--TStream------------------------------------------------------------------}
{  Open -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                 }
{---------------------------------------------------------------------------}
procedure TStream.Open(OpenMode: Word);
begin { Abstract method }
end;

{--TStream------------------------------------------------------------------}
{  Error -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                }
{---------------------------------------------------------------------------}
procedure TStream.Error(Code, Info: Integer);
type
  TErrorProc = procedure(var s: TStream);
begin
  Status := Code; { Hold error code }
  ErrorInfo := Info; { Hold error info }
  if (StreamError <> nil) then
    TErrorProc(StreamError)(Self); { Call error ptr }
end;

{--TStream------------------------------------------------------------------}
{  Read -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                 }
{---------------------------------------------------------------------------}
procedure TStream.Read(var Buf; Count: Sw_Word);
begin
  Abstract; { Abstract error }
end;

{--TStream------------------------------------------------------------------}
{  Write -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB                }
{---------------------------------------------------------------------------}
procedure TStream.Write(var Buf; Count: Sw_Word);
begin
  Abstract; { Abstract error }
end;

{--TStream------------------------------------------------------------------}
{  CopyFrom -> Platforms DOS/DPMI/WIN/OS2 - Checked 10May96 LdB             }
{---------------------------------------------------------------------------}
procedure TStream.CopyFrom(var s: TStream; Count: LongInt);
var
  w: Word; Buffer: array[0..1023] of Byte;
begin
  while (Count > 0) do
    begin
      if (Count > SizeOf(Buffer)) then { To much data }
        w := SizeOf(Buffer)
      else
        w := Count; { Size to transfer }
      s.Read(Buffer, w); { Read from stream }
      Write(Buffer, w); { Write to stream }
      Dec(Count, w); { Dec write count }
    end;
end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                       TCollection OBJECT METHODS                          }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

{--TCollection--------------------------------------------------------------}
{  Init -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                 }
{---------------------------------------------------------------------------}
constructor TCollection.Init(ALimit, ADelta: Sw_Integer);
begin
  inherited Init; { Call ancestor }
  Delta := ADelta; { Set increment }
  SetLimit(ALimit); { Set limit }
end;

{--TCollection--------------------------------------------------------------}
{  Load -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                 }
{---------------------------------------------------------------------------}

(* RJ
constructor TCollection.Load (var s: TStream);
var
  c, i              : Sw_Integer;
begin
  if s.TPCompatible then
    begin
      { I ignore endianness issues here. If endianness is different,
        you can't expect binary compatible resources anyway. }
      Count := 0; s.Read (Count, SizeOf (Word));
      Limit := 0; s.Read (Limit, SizeOf (Word));
      Delta := 0; s.Read (Delta, SizeOf (Word))
    end
  else
    begin
      s.Read (Count, SizeOf (Count));   { Read count }
      s.Read (Limit, SizeOf (Limit));   { Read limit }
      s.Read (Delta, SizeOf (Delta));   { Read delta }
    end;
  Items := nil;                         { Clear item pointer }
  c := Count;                           { Hold count }
  i := Limit;                           { Hold limit }
  Count := 0;                           { Clear count }
  Limit := 0;                           { Clear limit }
  SetLimit (i);                         { Set requested limit }
  Count := c;                           { Set count }
  for i := 0 to c - 1 do AtPut (i, GetItem (s)); { Get each item }
end;
*)

{--TCollection--------------------------------------------------------------}
{  Done -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                 }
{---------------------------------------------------------------------------}
destructor TCollection.Done;
begin
  FreeAll; { Free all items }
  SetLimit(0); { Release all memory }
end;

{--TCollection--------------------------------------------------------------}
{  At -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                   }
{---------------------------------------------------------------------------}
function TCollection.At(Index: Sw_Integer): Pointer;
begin
  if (Index < 0) or (Index >= Count) then
    begin { Invalid index }
      Error(coIndexError, Index); { Call error }
      At := nil; { Return nil }
    end
  else
    At := Items^[Index]; { Return item }
end;

{--TCollection--------------------------------------------------------------}
{  IndexOf -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB              }
{---------------------------------------------------------------------------}
function TCollection.IndexOf(Item: Pointer): Sw_Integer;
var
  i: Sw_Integer;
begin
  if (Count > 0) then
    begin { Count is positive }
      for i := 0 to Count - 1 do { For each item }
        if (Items^[i] = Item) then
          begin { Look for match }
            IndexOf := i; { Return index }
            Exit; { Now exit }
          end;
    end;
  IndexOf := -1; { Return index }
end;

{--TCollection--------------------------------------------------------------}
{  GetItem -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB              }
{---------------------------------------------------------------------------}

(* RJ
function TCollection.GetItem (var s: TStream): Pointer;
begin
  GetItem := s.Get;                     { Item off stream }
end;
*)

{--TCollection--------------------------------------------------------------}
{  LastThat -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB             }
{---------------------------------------------------------------------------}

(*

function TCollection.LastThat (test: Pointer): Pointer;
var
  i                 : LongInt;

begin
  for i := Count downto 1 do
    begin                               { Down from last item }
      if Boolean (Byte (LongInt (CallPointerLocal (test, PreviousFramePointer, Items^[i - 1])))) then
        begin                           { Test each item }
          LastThat := Items^[i - 1];    { Return item }
          Exit;                         { Now exit }
        end;
    end;
  LastThat := nil;                      { None passed test }
end;

*)

{--TCollection--------------------------------------------------------------}
{  FirstThat -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB            }
{---------------------------------------------------------------------------}

(* RJ
function TCollection.FirstThat (test: Pointer): Pointer;
var
  i                 : LongInt;
begin
  for i := 1 to Count do
    begin                               { Up from first item }
      if Boolean (Byte (LongInt (CallPointerLocal (test, PreviousFramePointer, Items^[i - 1])))) then
        begin                           { Test each item }
          FirstThat := Items^[i - 1];   { Return item }
          Exit;                         { Now exit }
        end;
    end;
  FirstThat := nil;                     { None passed test }
end;

*)

{--TCollection--------------------------------------------------------------}
{  Pack -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                 }
{---------------------------------------------------------------------------}
procedure TCollection.Pack;
var
  i, j: Sw_Integer;
begin
  i := 0; { Initialize dest }
  j := 0; { Intialize test }
  while (i < Count) and (j < Limit) do
    begin { Check fully packed }
      if (Items^[j] <> nil) then
        begin { Found a valid item }
          if (i <> j) then
            begin
              Items^[i] := Items^[j]; { Transfer item }
              Items^[j] := nil; { Now clear old item }
            end;
          Inc(i); { One item packed }
        end;
      Inc(j); { Next item to test }
    end;
  if (i < Count) then Count := i; { New packed count }
end;

{--TCollection--------------------------------------------------------------}
{  FreeAll -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB              }
{---------------------------------------------------------------------------}
procedure TCollection.FreeAll;
var
  i: Sw_Integer;
begin
  for i := Count - 1 downto 0 do
    FreeItem(At(i));
  Count := 0; { Clear item count }
end;

{--TCollection--------------------------------------------------------------}
{  DeleteAll -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB            }
{---------------------------------------------------------------------------}
procedure TCollection.DeleteAll;
begin
  Count := 0; { Clear item count }
end;

{--TCollection--------------------------------------------------------------}
{  Free -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                 }
{---------------------------------------------------------------------------}
procedure TCollection.Free(Item: Pointer);
begin
  Delete(Item); { Delete from list }
  FreeItem(Item); { Free the item }
end;

{--TCollection--------------------------------------------------------------}
{  Insert -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB               }
{---------------------------------------------------------------------------}
procedure TCollection.Insert(Item: Pointer);
begin
  AtInsert(Count, Item); { Insert item }
end;

{--TCollection--------------------------------------------------------------}
{  Delete -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB               }
{---------------------------------------------------------------------------}
procedure TCollection.Delete(Item: Pointer);
begin
  AtDelete(IndexOf(Item)); { Delete from list }
end;

{--TCollection--------------------------------------------------------------}
{  AtFree -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB               }
{---------------------------------------------------------------------------}
procedure TCollection.AtFree(Index: Sw_Integer);
var
  Item: Pointer;
begin
  Item := At(Index); { Retreive item ptr }
  AtDelete(Index); { Delete item }
  FreeItem(Item); { Free the item }
end;

{--TCollection--------------------------------------------------------------}
{  FreeItem -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB             }
{---------------------------------------------------------------------------}
procedure TCollection.FreeItem(Item: Pointer);
var
  p: PObject;
begin
  p := PObject(Item); { Convert pointer }
  if (p <> nil) then Dispose(p, Done); { Dispose of object }
end;

{--TCollection--------------------------------------------------------------}
{  AtDelete -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB             }
{---------------------------------------------------------------------------}
procedure TCollection.AtDelete(Index: Sw_Integer);
begin
  if (Index >= 0) and (Index < Count) then
    begin { Valid index }
      Dec(Count); { One less item }
      if (Count > Index) then
        Move(Items^[Index + 1],
          Items^[Index], (Count - Index) * SizeOf(Pointer)); { Shuffle items down }
    end
  else
    Error(coIndexError, Index); { Index error }
end;

{--TCollection--------------------------------------------------------------}
{  ForEach -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB              }
{---------------------------------------------------------------------------}

(* RJ

procedure TCollection.forEach (Action: Pointer);
var
  i                 : LongInt;
begin
  for i := 1 to Count do                { Up from first item }
    CallPointerLocal (Action, PreviousFramePointer, Items^[i - 1]); { Call with each item }
end;

*)

{--TCollection--------------------------------------------------------------}
{  SetLimit -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB             }
{---------------------------------------------------------------------------}
procedure TCollection.SetLimit(ALimit: Sw_Integer);
var
  AItems: PItemList;
begin
  if (ALimit < Count) then
    ALimit := Count;
  if (ALimit > MaxCollectionSize) then
    ALimit := MaxCollectionSize;
  if (ALimit <> Limit) then
    begin
      if (ALimit = 0) then
        AItems := nil
      else
        begin
          GetMem(AItems, ALimit * SizeOf(Pointer));
          if (AItems <> nil) then
            FillChar(AItems^, ALimit * SizeOf(Pointer), #0);
        end;
      if (AItems <> nil) or (ALimit = 0) then
        begin
          if (AItems <> nil) and (Items <> nil) then
            Move(Items^, AItems^, Count * SizeOf(Pointer));
          if (Limit <> 0) and (Items <> nil) then
            FreeMem(Items, Limit * SizeOf(Pointer));
        end;
      Items := AItems;
      Limit := ALimit;
    end;
end;

{--TCollection--------------------------------------------------------------}
{  Error -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                }
{---------------------------------------------------------------------------}
procedure TCollection.Error(Code, Info: Integer);
begin
  RunError(212 - Code); { Run error }
end;

{--TCollection--------------------------------------------------------------}
{  AtPut -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                }
{---------------------------------------------------------------------------}
procedure TCollection.AtPut(Index: Sw_Integer; Item: Pointer);
begin
  if (Index >= 0) and (Index < Count) then { Index valid }
    Items^[Index] := Item { Put item in index }
  else
    Error(coIndexError, Index); { Index error }
end;

{--TCollection--------------------------------------------------------------}
{  AtInsert -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB             }
{---------------------------------------------------------------------------}
procedure TCollection.AtInsert(Index: Sw_Integer; Item: Pointer);
var
  i: Sw_Integer;
begin
  if (Index >= 0) and (Index <= Count) then
    begin { Valid index }
      if (Count = Limit) then SetLimit(Limit + Delta); { Expand size if able }
      if (Limit > Count) then
        begin
          if (Index < Count) then
            begin { Not last item }
              for i := Count - 1 downto Index do { Start from back }
                Items^[i + 1] := Items^[i]; { Move each item }
            end;
          Items^[Index] := Item; { Put item in list }
          Inc(Count); { Inc count }
        end
      else
        Error(coOverflow, Index); { Expand failed }
    end
  else
    Error(coIndexError, Index); { Index error }
end;

{--TCollection--------------------------------------------------------------}
{  Store -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                }
{---------------------------------------------------------------------------}

(* RJ
procedure TCollection.STORE (var s: TStream);
var
  LimitWord, DeltaWord: Word;

  procedure DoPutItem (p: Pointer); {$IFNDEF FPC}far; {$ENDIF}
  begin
    PutItem (s, p);                     { Put item on stream }
  end;

begin
  if s.TPCompatible then
    begin
      { Check if it is safe to write in TP-compatible stream.
        If Count is too big, signal an error.
        If Limit or Delta are too big, write smaller values. }
      if (Count > MaxTPCompatibleCollectionSize)
        then s.Error (stWriteError, 0)
      else
        begin
          s.Write (Count, SizeOf (Word));
          if Limit > MaxTPCompatibleCollectionSize
            then LimitWord := MaxTPCompatibleCollectionSize
          else LimitWord := Limit;
          s.Write (LimitWord, SizeOf (Word));
          if Delta > MaxTPCompatibleCollectionSize
            then DeltaWord := MaxTPCompatibleCollectionSize
          else DeltaWord := Delta;
          s.Write (DeltaWord, SizeOf (Word));
        end
    end
  else
    begin
      s.Write (Count, SizeOf (Count));  { Write count }
      s.Write (Limit, SizeOf (Limit));  { Write limit }
      s.Write (Delta, SizeOf (Delta));  { Write delta }
    end;
  forEach (@DoPutItem);                 { Each item to stream }
end;
*)

{--TCollection--------------------------------------------------------------}
{  PutItem -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB              }
{---------------------------------------------------------------------------}

(* RJ
procedure TCollection.PutItem (var s: TStream; Item: Pointer);
begin
  s.Put (Item);                         { Put item on stream }
end;
*)

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                       TSortedCollection OBJECT METHODS                    }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

{--TSortedCollection--------------------------------------------------------}
{  Init -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                 }
{---------------------------------------------------------------------------}
constructor TSortedCollection.Init(ALimit, ADelta: Sw_Integer);
begin
  inherited Init(ALimit, ADelta); { Call ancestor }
  Duplicates := False; { Clear flag }
end;

{--TSortedCollection--------------------------------------------------------}
{  Load -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                 }
{---------------------------------------------------------------------------}

(* RJ
constructor TSortedCollection.Load (var s: TStream);
begin
  inherited Load (s);                   { Call ancestor }
  s.Read (Duplicates, SizeOf (Duplicates)); { Read duplicate flag }
end;
*)

{--TSortedCollection--------------------------------------------------------}
{  KeyOf -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                }
{---------------------------------------------------------------------------}
function TSortedCollection.KeyOf(Item: Pointer): Pointer;
begin
  KeyOf := Item; { Return item as key }
end;

{--TSortedCollection--------------------------------------------------------}
{  IndexOf -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB              }
{---------------------------------------------------------------------------}
function TSortedCollection.IndexOf(Item: Pointer): Sw_Integer;
var
  i, j: Sw_Integer;
begin
  j := -1; { Preset result }
  if Search(KeyOf(Item), i) then
    begin { Search for item }
      if Duplicates then { Duplicates allowed }
        while (i < Count) and (Item <> Items^[i]) do
          Inc(i); { Count duplicates }
      if (i < Count) then j := i; { Index result }
    end;
  IndexOf := j; { Return result }
end;

{--TSortedCollection--------------------------------------------------------}
{  Compare -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB              }
{---------------------------------------------------------------------------}
function TSortedCollection.Compare(Key1, Key2: Pointer): Sw_Integer;
begin
  Abstract; { Abstract method }
  Compare := 0;
end;

{--TSortedCollection--------------------------------------------------------}
{  Search -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB               }
{---------------------------------------------------------------------------}
function TSortedCollection.Search(Key: Pointer; var Index: Sw_Integer): Boolean;
var
  l, h, i, c: Sw_Integer;
begin
  Search := False; { Preset failure }
  l := 0; { Start count }
  h := Count - 1; { End count }
  while (l <= h) do
    begin
      i := (l + h) shr 1; { Mid point }
      c := Compare(KeyOf(Items^[i]), Key); { Compare with key }
      if (c < 0) then
        l := i + 1
      else
        begin { Item to left }
          h := i - 1; { Item to right }
          if c = 0 then
            begin { Item match found }
              Search := True; { Result true }
              if not Duplicates then l := i; { Force kick out }
            end;
        end;
    end;
  Index := l; { Return result }
end;

{--TSortedCollection--------------------------------------------------------}
{  Insert -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB               }
{---------------------------------------------------------------------------}
procedure TSortedCollection.Insert(Item: Pointer);
var
  i: Sw_Integer;
begin
  if not Search(KeyOf(Item), i) or Duplicates then { Item valid }
    AtInsert(i, Item); { Insert the item }
end;

{--TSortedCollection--------------------------------------------------------}
{  Store -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB                }
{---------------------------------------------------------------------------}

(*
procedure TSortedCollection.STORE (var s: TStream);
begin
  TCollection.STORE (s);                { Call ancestor }
  s.Write (Duplicates, SizeOf (Duplicates)); { Write duplicate flag }
end;
*)

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                     TStringCollection OBJECT METHODS                      }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

{--TStringCollection--------------------------------------------------------}
{  GetItem -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB              }
{---------------------------------------------------------------------------}
function TStringCollection.GetItem(var s: TStream): Pointer;
begin
  GetItem := s.ReadStr; { Get new item }
end;

{--TStringCollection--------------------------------------------------------}
{  Compare -> Platforms DOS/DPMI/WIN/OS2 - Checked 21Aug97 LdB              }
{---------------------------------------------------------------------------}
function TStringCollection.Compare(Key1, Key2: Pointer): Sw_Integer;
var
  i, j: Sw_Integer; p1, p2: PString;
begin
  p1 := PString(Key1); { String 1 pointer }
  p2 := PString(Key2); { String 2 pointer }
  if (Length(p1^) < Length(p2^)) then
    j := Length(p1^)
  else
    j := Length(p2^); { Shortest length }
  i := 1; { First character }
  while (i < j) and (p1^[i] = p2^[i]) do
    Inc(i); { Scan till fail }
  if (i = j) then
    begin { Possible match }
      { * REMARK * - Bug fix   21 August 1997 }
      if (p1^[i] < p2^[i]) then
        Compare := -1
      else { String1 < String2 }
        if (p1^[i] > p2^[i]) then
          Compare := 1
        else { String1 > String2 }
          if (Length(p1^) > Length(p2^)) then
            Compare := 1 { String1 > String2 }
          else
            if (Length(p1^) < Length(p2^)) then { String1 < String2 }
              Compare := -1
            else
              Compare := 0; { String1 = String2 }
      { * REMARK END * - Leon de Boer }
    end
  else
    if (p1^[i] < p2^[i]) then
      Compare := -1 { String1 < String2 }
    else
      Compare := 1; { String1 > String2 }
end;

{--TStringCollection--------------------------------------------------------}
{  FreeItem -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB             }
{---------------------------------------------------------------------------}
procedure TStringCollection.FreeItem(Item: Pointer);
begin
  DisposeStr(Item); { Dispose item }
end;

{--TStringCollection--------------------------------------------------------}
{  PutItem -> Platforms DOS/DPMI/WIN/OS2 - Checked 22May96 LdB              }
{---------------------------------------------------------------------------}
procedure TStringCollection.PutItem(var s: TStream; Item: Pointer);
begin
  s.WriteStr(Item); { Write string }
end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                   TUnSortedStrCollection OBJECT METHODS                   }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

{--TUnSortedCollection------------------------------------------------------}
{  Insert -> Platforms DOS/DPMI/WIN/OS2 - Checked 23May96 LdB               }
{---------------------------------------------------------------------------}
procedure TUnSortedStringCollection.Insert(Item: Pointer);
begin
  AtInsert(Count, Item); { Insert - NO sorting }
end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                           TResourceItem RECORD                            }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
type
  TResourceItem = packed record
    Posn: LongInt; { Resource position }
    Size: LongInt; { Resource size }
    Key: string; { Resource key }
  end;
  PResourceItem = ^TResourceItem;

  {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  {                  PRIVATE RESOURCE MANAGER CONSTANTS                       }
  {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
const
  RStreamMagic: LongInt = $52504246; { 'FBPR' }
  RStreamBackLink: LongInt = $4C424246; { 'FBBL' }

  {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  {                    PRIVATE RESOURCE MANAGER TYPES                         }
  {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
type
  {$IFDEF NewExeFormat} { New EXE format }
  TExeHeader = packed record
    eHdrSize: Word;
    eMinAbove: Word;
    eMaxAbove: Word;
    eInitSS: Word;
    eInitSP: Word;
    eCheckSum: Word;
    eInitPC: Word;
    eInitCS: Word;
    eRelocOfs: Word;
    eOvlyNum: Word;
    eRelocTab: Word;
    eSpace: array[1..30] of Byte;
    eNewHeader: Word;
  end;
  {$ENDIF}

  THeader = packed record
    Signature: Word;
    case Integer of
      0: (
        LastCount: Word;
        PageCount: Word;
        ReloCount: Word);
      1: (
        InfoType: Word;
        InfoSize: LongInt);
  end;

  (*

    {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
    {                       TResourceFile OBJECT METHODS                        }
    {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

    {--TResourceFile------------------------------------------------------------}
    {  Init -> Platforms DOS/DPMI/WIN/OS2 - Checked 18Jun96 LdB                 }
    {---------------------------------------------------------------------------}
  constructor TResourceFile.Init (AStream: PStream);
  var
    Found, Stop       : Boolean; Header: THeader;
    {$IFDEF NewExeFormat}ExeHeader: TExeHeader; {$ENDIF}
  begin
    TObject.Init;                         { Initialize object }
    Found := False;                       { Preset false }
    if (AStream <> nil) then
      begin
        Stream := AStream;                { Hold stream }
        BasePos := Stream^.GetPos;        { Get position }
        repeat
          Stop := True;                   { Preset stop }
          if (BasePos <= Stream^.GetSize - SizeOf (THeader)) then
            begin                         { Valid file header }
              Stream^.Seek (BasePos);     { Seek to position }
              Stream^.Read (Header, SizeOf (THeader)); { Read header }
              case Header.Signature of
                {$IFDEF NewExeFormat}     { New format file }
                $5A4D:
                  begin
                    Stream^.Read (ExeHeader, SizeOf (TExeHeader));
                    BasePos := ExeHeader.eNewHeader; { Hold position }
                    Stop := False;        { Clear stop flag }
                  end;
                $454E:
                  begin
                    BasePos := Stream^.GetSize - 8; { Hold position }
                    Stop := False;        { Clear stop flag }
                  end;
                $4246:
                  begin
                    Stop := False;        { Clear stop flag }
                    case Header.InfoType of
                      $5250:
                        begin             { Found Resource }
                          Found := True;  { Found flag is true }
                          Stop := True;   { Set stop flag }
                        end;
                      $4C42: Dec (BasePos, Header.InfoSize - 8); { Found BackLink }
                      $4648: Dec (BasePos, SizeOf (THeader) * 2); { Found HelpFile }
                    else
                      Stop := True;       { Set stop flag }
                    end;
                  end;
                $424E:
                  if Header.InfoType = $3230 { Found Debug Info } then
                    begin
                      Dec (BasePos, Header.InfoSize); { Adjust position }
                      Stop := False;      { Clear stop flag }
                    end;
                {$ELSE}
                $5A4D:
                  begin
                    Inc (BasePos, LongInt (Header.PageCount) * 512
                      - (-Header.LastCount and 511)); { Calc position }
                    Stop := False;        { Clear stop flag }
                  end;
                $4246:
                  if Header.InfoType = $5250 then { Header was found }
                    Found := True
                  else
                    begin
                      Inc (BasePos, Header.InfoSize + 8); { Adjust position }
                      Stop := False;      { Clear stop flag }
                    end;
                {$ENDIF}
              end;
            end;
        until Stop;                       { Until flag is set }
      end;
    if Found then
      begin                               { Resource was found }
        Stream^.Seek (BasePos + SizeOf (LongInt) * 2); { Seek to position }
        Stream^.Read (IndexPos, SizeOf (LongInt)); { Read index position }
        Stream^.Seek (BasePos + IndexPos); { Seek to resource }
        Index.Load (Stream^);             { Load resource }
      end
    else
      begin
        IndexPos := SizeOf (LongInt) * 3; { Set index position }
        Index.Init (0, 8);                { Set index }
      end;
  end;

  {--TResourceFile------------------------------------------------------------}
  {  Done -> Platforms DOS/DPMI/WIN/OS2 - Checked 18Jun96 LdB                 }
  {---------------------------------------------------------------------------}
  destructor TResourceFile.Done;
  begin
    Flush;                                { Flush the file }
    Index.Done;                           { Dispose of index }
    if (Stream <> nil) then Dispose (Stream, Done); { Dispose of stream }
  end;

  {--TResourceFile------------------------------------------------------------}
  {  Count -> Platforms DOS/DPMI/WIN/OS2 - Checked 18Jun96 LdB                }
  {---------------------------------------------------------------------------}
  function TResourceFile.Count: Sw_Integer;
  begin
    Count := Index.Count;                 { Return index count }
  end;

  {--TResourceFile------------------------------------------------------------}
  {  KeyAt -> Platforms DOS/DPMI/WIN/OS2 - Checked 18Jun96 LdB                }
  {---------------------------------------------------------------------------}
  function TResourceFile.KeyAt (i: Sw_Integer): string;
  begin
    KeyAt := PResourceItem (Index.At (i))^.Key; { Return key }
  end;

  {--TResourceFile------------------------------------------------------------}
  {  Get -> Platforms DOS/DPMI/WIN/OS2 - Checked 18Jun96 LdB                  }
  {---------------------------------------------------------------------------}

  function TResourceFile.Get (Key: string): PObject;
  var
    i                 : Sw_Integer;
  begin
    if (Stream = nil) or (not Index.Search (@Key, i)) { No match on key }
    then Get := nil else
      begin
        Stream^.Seek (BasePos +
          PResourceItem (Index.At (i))^.Posn); { Seek to position }
        Get := Stream^.Get;               { Get item }
      end;
  end;

  {--TResourceFile------------------------------------------------------------}
  {  SwitchTo -> Platforms DOS/DPMI/WIN/OS2 - Checked 18Jun96 LdB             }
  {---------------------------------------------------------------------------}

  function TResourceFile.SwitchTo (AStream: PStream; Pack: Boolean): PStream;
  var
    NewBasePos        : LongInt;

    procedure DoCopyResource (Item: PResourceItem); {$IFNDEF FPC}far; {$ENDIF}
    begin
      Stream^.Seek (BasePos + Item^.Posn); { Move stream position }
      Item^.Posn := AStream^.GetPos - NewBasePos; { Hold new position }
      AStream^.CopyFrom (Stream^, Item^.Size); { Copy the item }
    end;

  begin
    SwitchTo := Stream;                   { Preset return }
    if (AStream <> nil) and (Stream <> nil) then
      begin                               { Both streams valid }
        NewBasePos := AStream^.GetPos;    { Get position }
        if Pack then
          begin
            AStream^.Seek (NewBasePos + SizeOf (LongInt) * 3); { Seek to position }
            Index.forEach (@DoCopyResource); { Copy each resource }
            IndexPos := AStream^.GetPos - NewBasePos; { Hold index position }
          end else
          begin
            Stream^.Seek (BasePos);       { Seek to position }
            AStream^.CopyFrom (Stream^, IndexPos); { Copy the resource }
          end;
        Stream := AStream;                { Hold new stream }
        BasePos := NewBasePos;            { New base position }
        Modified := True;                 { Set modified flag }
      end;
  end;

  {--TResourceFile------------------------------------------------------------}
  {  Flush -> Platforms DOS/DPMI/WIN/OS2 - Checked 18Jun96 LdB                }
  {---------------------------------------------------------------------------}
  procedure TResourceFile.Flush;
  var
    ResSize           : LongInt; LinkSize: LongInt;
  begin
    if (Modified) and (Stream <> nil) then
      begin                               { We have modification }
        Stream^.Seek (BasePos + IndexPos); { Seek to position }
        Index.STORE (Stream^);            { Store the item }
        ResSize := Stream^.GetPos - BasePos; { Hold position }
        LinkSize := ResSize + SizeOf (LongInt) * 2; { Hold link size }
        Stream^.Write (RStreamBackLink, SizeOf (LongInt)); { Write link back }
        Stream^.Write (LinkSize, SizeOf (LongInt)); { Write link size }
        Stream^.Seek (BasePos);           { Move stream position }
        Stream^.Write (RStreamMagic, SizeOf (LongInt)); { Write number }
        Stream^.Write (ResSize, SizeOf (LongInt)); { Write record size }
        Stream^.Write (IndexPos, SizeOf (LongInt)); { Write index position }
        Stream^.Flush;                    { Flush the stream }
      end;
    Modified := False;                    { Clear modified flag }
  end;

  {--TResourceFile------------------------------------------------------------}
  {  Delete -> Platforms DOS/DPMI/WIN/OS2 - Checked 18Jun96 LdB               }
  {---------------------------------------------------------------------------}
  procedure TResourceFile.Delete (Key: string);
  var
    i                 : Sw_Integer;
  begin
    if Index.Search (@Key, i) then
      begin                               { Search for key }
        Index.Free (Index.At (i));        { Delete from index }
        Modified := True;                 { Set modified flag }
      end;
  end;

  {--TResourceFile------------------------------------------------------------}
  {  Put -> Platforms DOS/DPMI/WIN/OS2 - Checked 18Jun96 LdB                  }
  {---------------------------------------------------------------------------}

  procedure TResourceFile.Put (Item: PObject; Key: string);
  var
    i                 : Sw_Integer; ts: string; p: PResourceItem;
  begin
    if (Stream = nil) then Exit;          { Stream not valid }
    if Index.Search (@Key, i) then p := Index.At (i) { Search for item }
    else
      begin
        GetMem (p, Length (Key) + (SizeOf (TResourceItem) -
          SizeOf (ts) + 1));              { Allocate memory }
        if (p <> nil) then
          begin
            p^.Key := Key;                { Store key }
            Index.AtInsert (i, p);        { Insert item }
          end;
      end;
    if (p <> nil) then
      begin
        p^.Posn := IndexPos;              { Set index position }
        Stream^.Seek (BasePos + IndexPos); { Seek file position }
        Stream^.Put (Item);               { Put item on stream }
        IndexPos := Stream^.GetPos - BasePos; { Hold index position }
        p^.Size := IndexPos - p^.Posn;    { Calc size }
        Modified := True;                 { Set modified flag }
      end;
  end;

  *)

  {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  {                          TStringList OBJECT METHODS                       }
  {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  {--TStringList--------------------------------------------------------------}
  {  Load -> Platforms DOS/DPMI/WIN/OS2 - Checked 30Jun97 LdB                 }
  {---------------------------------------------------------------------------}
constructor TStringList.Load(var s: TStream);
var
  Size: Word;
begin
  Stream := @s; { Hold stream pointer }
  s.Read(Size, SizeOf(Word)); { Read size }
  BasePos := s.GetPos; { Hold position }
  s.Seek(BasePos + Size); { Seek to position }
  s.Read(IndexSize, SizeOf(Integer)); { Read index size }
  GetMem(Index, IndexSize * SizeOf(TStrIndexRec)); { Allocate memory }
  s.Read(Index^, IndexSize * SizeOf(TStrIndexRec)); { Read indexes }
end;

{--TStringList--------------------------------------------------------------}
{  Done -> Platforms DOS/DPMI/WIN/OS2 - Checked 30Jun97 LdB                 }
{---------------------------------------------------------------------------}
destructor TStringList.Done;
begin
  FreeMem(Index, IndexSize * SizeOf(TStrIndexRec)); { Release memory }
end;

{--TStringList--------------------------------------------------------------}
{  Get -> Platforms DOS/DPMI/WIN/OS2 - Checked 30Jun97 LdB                  }
{---------------------------------------------------------------------------}
function TStringList.Get(Key: Sw_Word): string;
var
  i: Word; s: string;
begin
  s := ''; { Preset empty string }
  if (IndexSize > 0) then
    begin { We must have strings }
      i := 0; { First entry }
      while (i < IndexSize) and (s = '') do
        begin
          if ((Key - Index^[i].Key) < Index^[i].Count) { Diff less than count } then
            ReadStr(s, Index^[i].Offset,
              Key - Index^[i].Key); { Read the string }
          Inc(i); { Next entry }
        end;
    end;
  Get := s; { Return empty string }
end;

{***************************************************************************}
{                       TStringList PRIVATE METHODS                         }
{***************************************************************************}

{--TStringLis---------------------------------------------------------------}
{  ReadStr -> Platforms DOS/DPMI/WIN/OS2 - Checked 30Jun97 LdB              }
{---------------------------------------------------------------------------}
procedure TStringList.ReadStr(var s: string; Offset, Skip: Sw_Word);
begin
  Stream^.Seek(BasePos + Offset); { Seek to position }
  Inc(Skip); { Adjust skip }
  repeat
    Stream^.Read(s[0], 1); { Read string size }
    Stream^.Read(s[1], Ord(s[0])); { Read string data }
    Dec(Skip); { One string read }
  until (Skip = 0);
end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                         TStrListMaker OBJECT METHODS                      }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

{--TStrListMaker------------------------------------------------------------}
{  Init -> Platforms DOS/DPMI/WIN/OS2 - Checked 30Jun97 LdB                 }
{---------------------------------------------------------------------------}
constructor TStrListMaker.Init(AStrSize, AIndexSize: Sw_Word);
begin
  inherited Init; { Call ancestor }
  StrSize := AStrSize; { Hold size }
  IndexSize := AIndexSize; { Hold index size }
  GetMem(strings, AStrSize); { Allocate memory }
  GetMem(Index, AIndexSize * SizeOf(TStrIndexRec)); { Allocate memory }
end;

{--TStrListMaker------------------------------------------------------------}
{  Done -> Platforms DOS/DPMI/WIN/OS2 - Checked 30Jun97 LdB                 }
{---------------------------------------------------------------------------}
destructor TStrListMaker.Done;
begin
  FreeMem(Index, IndexSize * SizeOf(TStrIndexRec)); { Free index memory }
  FreeMem(strings, StrSize); { Free data memory }
end;

{--TStrListMaker------------------------------------------------------------}
{  Put -> Platforms DOS/DPMI/WIN/OS2 - Checked 30Jun97 LdB                  }
{---------------------------------------------------------------------------}
procedure TStrListMaker.Put(Key: Sw_Word; s: string);
begin
  if (cur.Count = 16) or (Key <> cur.Key + cur.Count) then CloseCurrent; { Close current }
  if (cur.Count = 0) then
    begin
      cur.Key := Key; { Set key }
      cur.Offset := StrPos; { Set offset }
    end;
  Inc(cur.Count); { Inc count }
  Move(s, strings^[StrPos], Length(s) + 1); { Move string data }
  Inc(StrPos, Length(s) + 1); { Adjust position }
end;

{--TStrListMaker------------------------------------------------------------}
{  Store -> Platforms DOS/DPMI/WIN/OS2 - Checked 30Jun97 LdB                }
{---------------------------------------------------------------------------}
procedure TStrListMaker.STORE(var s: TStream);
begin
  CloseCurrent; { Close all current }
  s.Write(StrPos, SizeOf(Word)); { Write position }
  s.Write(strings^, StrPos); { Write string data }
  s.Write(IndexPos, SizeOf(Word)); { Write index position }
  s.Write(Index^, IndexPos * SizeOf(TStrIndexRec)); { Write indexes }
end;

{***************************************************************************}
{                      TStrListMaker PRIVATE METHODS                        }
{***************************************************************************}

{--TStrListMaker------------------------------------------------------------}
{  CloseCurrent -> Platforms DOS/DPMI/WIN/OS2 - Checked 30Jun97 LdB         }
{---------------------------------------------------------------------------}
procedure TStrListMaker.CloseCurrent;
begin
  if (cur.Count <> 0) then
    begin
      Index^[IndexPos] := cur; { Hold index position }
      Inc(IndexPos); { Next index }
      cur.Count := 0; { Adjust count }
    end;
end;

{***************************************************************************}
{                            INTERFACE ROUTINES                             }
{***************************************************************************}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                    DYNAMIC STRING INTERFACE ROUTINES                      }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

{---------------------------------------------------------------------------}
{  NewStr -> Platforms DOS/DPMI/WINDOWS/OS2 - Checked 12Jun96 LdB           }
{---------------------------------------------------------------------------}
function NewStr(const s: string): PString;
// var
//  p                 : PString;
begin
  if (s = '') then
    Result := nil
  else
    begin { Return nil }
      GetMem(Result, Length(s) + 1); { Allocate memory }
      if (Result <> nil) then Result^ := s; { Hold string }
    end;
  //  Result := p;                          { Return result }
end;

{---------------------------------------------------------------------------}
{  DisposeStr -> Platforms DOS/DPMI/WINDOWS/OS2 - Checked 12Jun96 LdB       }
{---------------------------------------------------------------------------}
procedure DisposeStr(p: PString);
begin
  if (p <> nil) then FreeMem(p, Length(p^) + 1); { Release memory }
end;

procedure SetStr(var p: PString; const s: string);
begin
  if p <> nil then
    FreeMem(p, Length(p^) + 1);
  GetMem(p, Length(s) + 1);
  PString(p)^ := s
end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                        STREAM INTERFACE ROUTINES                          }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

{---------------------------------------------------------------------------}
{  Abstract -> Platforms DOS/DPMI/WINDOWS/OS2 - Checked 12Jun96 LdB         }
{---------------------------------------------------------------------------}
procedure Abstract;
begin
  RunError(211); { Abstract error }
end;

{---------------------------------------------------------------------------}
{  RegisterObjects -> Platforms DOS/DPMI/WINDOWS/OS2 - Checked 02Sep97 LdB  }
{---------------------------------------------------------------------------}
procedure RegisterObjects;
begin
  // RegisterType(RCollection);            { Register object }
  // RegisterType(RStringCollection);      { Register object }
  // RegisterType(RStrCollection);         { Register object }
end;

{---------------------------------------------------------------------------}
{  RegisterType -> Platforms DOS/DPMI/WINDOWS/OS2 - Checked 02Sep97 LdB     }
{---------------------------------------------------------------------------}
procedure RegisterType(var s: TStreamRec);
var
  p: PStreamRec;
begin
  p := StreamTypes; { Current reg list }
  while (p <> nil) and (p^.ObjType <> s.ObjType) do
    p := p^.Next; { Find end of chain }
  if (p = nil) and (s.ObjType <> 0) then
    begin { Valid end found }
      s.Next := StreamTypes; { Chain the list }
      StreamTypes := @s; { We are now first }
    end
  else
    RegisterError; { Register the error }
end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{                    GENERAL FUNCTION INTERFACE ROUTINES                    }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

{---------------------------------------------------------------------------}
{  LongMul -> Platforms DOS/DPMI/WINDOWS/OS2 - Checked 04Sep97 LdB          }
{---------------------------------------------------------------------------}
function LongMul(X, y: Integer): LongInt;
begin
  LongMul := LongInt(X * y);
end;
{---------------------------------------------------------------------------}
{  LongDiv -> Platforms DOS/DPMI/WINDOWS/OS2 - Checked 04Sep97 LdB          }
{---------------------------------------------------------------------------}
function LongDiv(X: LongInt; y: Integer): Integer;
begin
  LongDiv := Integer(X div y);
end;

end.

