{
    $Id: objinc.inc,v 1.6 2000/05/29 05:32:51 jonas Exp $
    This file is part of the Free Pascal run time library.
    Copyright (c) 1999-2000 by the Free Pascal development team.

    Includefile for objects.pp implementing OS-dependent file routines
    for WIN32

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************
}

const
  { REQUIRED TO PUT MANUALLY here, because of name conflicts in win32.inc }
  { flags for CreateFile }
  GENERIC_READ = $80000000;
  GENERIC_WRITE = $40000000;
  CREATE_New = 1;
  CREATE_ALWAYS = 2;
  OPEN_EXISTING = 3;
  OPEN_ALWAYS = 4;
  TRUNCATE_EXISTING = 5;

  FILE_ATTRIBUTE_ARCHIVE = 32;
  FILE_ATTRIBUTE_COMPRESSED = 2048;
  FILE_ATTRIBUTE_NORMAL = 128;
  FILE_ATTRIBUTE_DIRECTORY = 16;
  FILE_ATTRIBUTE_HIDDEN = 2;
  FILE_ATTRIBUTE_READONLY = 1;
  FILE_ATTRIBUTE_SYSTEM = 4;
  FILE_ATTRIBUTE_TEMPORARY = 256;

  { flags for SetFilePos }
  FILE_BEGIN = 0;
  FILE_CURRENT = 1;
  FILE_END = 2;

  { misc. functions }
function GetLastError: DWORD;
  external 'kernel32' Name 'GetLastError';

function WriteFile(fh: LongInt; Buf: Pointer; Len: LongInt; var Loaded: LongInt;
  overlap: Pointer): LongInt;
  external 'kernel32' Name 'WriteFile';
function ReadFile(fh: LongInt; Buf: Pointer; Len: LongInt; var Loaded: LongInt;
  overlap: Pointer): LongInt;
  external 'kernel32' Name 'ReadFile';
function CloseHandle(h: LongInt): LongInt;
  external 'kernel32' Name 'CloseHandle';
function DeleteFile(p: PChar): LongInt;
  external 'kernel32' Name 'DeleteFileA';
function MoveFile(Old, _new: PChar): LongInt;
  external 'kernel32' Name 'MoveFileA';
function SetFilePointer(l1, l2: LongInt; l3: Pointer; l4: LongInt): LongInt;
  external 'kernel32' Name 'SetFilePointer';
function GetFileSize(h: LongInt; p: Pointer): LongInt;
  external 'kernel32' Name 'GetFileSize';
function CreateFile(Name: Pointer; access, sharing: LongInt;
  security: Pointer; how, Attr, template: LongInt): LongInt;
  external 'kernel32' Name 'CreateFileA';
function SetEndOfFile(h: LongInt): Boolean;
  external 'kernel32' Name 'SetEndOfFile';
function GetFileType(Handle: DWORD): DWORD;
  external 'kernel32' Name 'GetFileType';

{---------------------------------------------------------------------------}
{  FileClose -> Platforms WIN32            - Not checked                    }
{---------------------------------------------------------------------------}
function FileClose(Handle: THandle): Word;
begin
  CloseHandle(Handle);
  FileClose := 0;
end;

{---------------------------------------------------------------------------}
{  FileOpen -> Platforms WIN32            - Tested MVC                      }
{  Returns 0 on failure                                                     }
{---------------------------------------------------------------------------}

function FileOpen(var FileName: AsciiZ; Mode: Word): THandle;
var
  oflags, cd: DWORD;
  AHandle: LongInt;
begin
  { On opening reset error code }
  DosStreamError := 0;
  if Mode = stCreate then
    begin
      cd := CREATE_ALWAYS;
      oflags := GENERIC_WRITE or GENERIC_READ;
    end
  else
    begin
      cd := OPEN_EXISTING;
      { convert filemode to filerec modes }
      case (Mode and 3) of
        1: oflags := GENERIC_WRITE;
        2: oflags := GENERIC_WRITE or GENERIC_READ;
      else
        // Cover (Mode and 3) = 0 here
        oflags := GENERIC_READ;
      end;
    end;

  AHandle := CreateFile(Pointer(@FileName), oflags, 0, nil, cd, FILE_ATTRIBUTE_NORMAL, 0);
  if AHandle = -1 then
    begin
      FileOpen := 0;
      DosStreamError := Word(GetLastError);
    end
  else
    FileOpen := AHandle;
end;

{***************************************************************************}
{  DosSetFilePtr -> Platforms WIN32        - Tested MVC                     }
{***************************************************************************}
function SetFilePos(Handle: THandle; Pos: LongInt; MoveType: Word;
  var Actual: LongInt): Word;
begin
  { WARNING WIN32 CURRECTLY HAS THE SAME SEEK MODE AS MSDOS    }
  { if this changes don't forget to change and check the flags }
  { accordingly.                                               }
  Actual := SetFilePointer(Handle, Pos, nil, MoveType);
  if Actual = -1 then
    DosStreamError := Word(GetLastError);
  SetFilePos := DosStreamError; { Return any error }
end;

{---------------------------------------------------------------------------}
{  FileRead -> Platforms WIN32            - Tested MVC                      }
{---------------------------------------------------------------------------}
function FileRead(Handle: THandle; var Buf; Count: Sw_Word;
  var Actual: Sw_Word): Word;

var Res: LongInt;

begin
  Res := 0;
  if ReadFile(Handle, Pointer(@Buf), Count, Res, nil) = 0 then
    DosStreamError := Word(GetLastError);
  Actual := Res;
  FileRead := DosStreamError;
end;

{---------------------------------------------------------------------------}
{  FileWrite -> Platforms WIN32            - Not Checked                    }
{---------------------------------------------------------------------------}
function FileWrite(Handle: THandle; var Buf; Count: Sw_Word; var Actual: Sw_Word): Word;
begin
  if WriteFile(Handle, Pointer(@Buf), Count, Actual, nil) = 0 then
    begin
      DosStreamError := Word(GetLastError);
    end;
  FileWrite := DosStreamError;
end;

{---------------------------------------------------------------------------}
{  SetFileSize -> Platforms DOS          - Not Checked                      }
{---------------------------------------------------------------------------}
function SetFileSize(Handle: THandle; FileSize: LongInt): Word;
var Actual, Buf: LongInt;
begin
  SetFilePos(Handle, FileSize, 0, Actual);
  if (Actual = FileSize) then
    begin
      Actual := FileWrite(Handle, Buf, 0, Actual); { Truncate the file }
      if (Actual <> -1) then
        SetFileSize := 0
      else
        SetFileSize := 103; { File truncate error }
    end
  else
    SetFileSize := 103; { File truncate error }
end;

{
  $Log: objinc.inc,v $
  Revision 1.6  2000/05/29 05:32:51  jonas
    * should compile again

  Revision 1.5  2000/02/09 16:59:34  peter
    * truncated log

  Revision 1.4  2000/01/07 16:41:52  daniel
    * copyright 2000

  Revision 1.3  2000/01/07 16:32:34  daniel
    * copyright 2000 added

  Revision 1.2  1999/08/24 08:44:43  michael
  + Fixed setfilepos and openfile

}

